from pymiere.core import _format_object_to_py, _format_object_to_es, _eval_script_returning_object
from pymiere.objects.premiere_objects import *


def _eval_on_global_object(extend_property):
    result = _eval_script_returning_object(extend_property)
    if result == "undefined":
        return None
    # if it's an object search if class exists and return objects creation arguments
    if isinstance(result, dict) and "isObject" in result and result["isObject"] is True:
        # create key word argument list to create the object
        kwargs = dict(pymiere_id=result["pymiereId"])
        return kwargs
    return result


class StartVars(object):
    def __init__(self):
        super(StartVars, self).__init__()

    # ----- PROPERTIES -----
    @property
    def NaN(self):
        return _eval_on_global_object('NaN')
    @NaN.setter
    def NaN(self, NaN):
        PymiereBaseObject._check_type(NaN, float, 'StartVars.NaN')
        _eval_on_global_object("NaN = {}".format(_format_object_to_es(NaN)))

    @property
    def Infinity(self):
        return _eval_on_global_object('Infinity')
    @Infinity.setter
    def Infinity(self, Infinity):
        PymiereBaseObject._check_type(Infinity, float, 'StartVars.Infinity')
        _eval_on_global_object("Infinity = {}".format(_format_object_to_es(Infinity)))

    @property
    def undefined(self):
        return _eval_on_global_object('undefined')
    @undefined.setter
    def undefined(self, undefined):
        PymiereBaseObject._check_type(undefined, None, 'StartVars.undefined')
        _eval_on_global_object("undefined = {}".format(_format_object_to_es(undefined)))

    @property
    def qe(self):
        return _format_object_to_py(_eval_script_returning_object('qe'))
    @qe.setter
    def qe(self, qe):
        raise AttributeError("Attribute 'qe' is read-only")

    """ The application object """
    @property
    def app(self):
        return Application(**_eval_script_returning_object('app', as_kwargs=True))
    @app.setter
    def app(self, app):
        raise AttributeError("Attribute 'app' is read-only")

    @property
    def apps(self):
        return Array(**_eval_script_returning_object('apps', as_kwargs=True))

    @apps.setter
    def apps(self, apps):
        raise AttributeError("Attribute 'apps' is read-only")

    @property
    def document(self):
        return Document(**_eval_script_returning_object('document', as_kwargs=True))
    @document.setter
    def document(self, document):
        raise AttributeError("Attribute 'document' is read-only")

    @property
    def ProjectItemType(self):
        return ProjectItemType(**_eval_script_returning_object('ProjectItemType', as_kwargs=True))
    @ProjectItemType.setter
    def ProjectItemType(self, ProjectItemType):
        raise AttributeError("Attribute 'ProjectItemType' is read-only")

    @property
    def ScratchDiskType(self):
        return ScratchDiskType(**_eval_script_returning_object('ScratchDiskType', as_kwargs=True))
    @ScratchDiskType.setter
    def ScratchDiskType(self, ScratchDiskType):
        raise AttributeError("Attribute 'ScratchDiskType' is read-only")

    @property
    def RegisteredDirectories(self):
        return RegisteredDirectories(**_eval_script_returning_object('RegisteredDirectories', as_kwargs=True))
    @RegisteredDirectories.setter
    def RegisteredDirectories(self, RegisteredDirectories):
        raise AttributeError("Attribute 'RegisteredDirectories' is read-only")

    @property
    def UtilityFunctions(self):
        return UtilityFunctions(**_eval_script_returning_object('UtilityFunctions', as_kwargs=True))
    @UtilityFunctions.setter
    def UtilityFunctions(self, UtilityFunctions):
        raise AttributeError("Attribute 'UtilityFunctions' is read-only")

    @property
    def Dollar(self):
        return Dollar(**_eval_script_returning_object('$', as_kwargs=True))
    @Dollar.setter
    def Dollar(self, Dollar):
        raise AttributeError("Attribute 'Dollar' is read-only")

    @property
    def Math(self):
        return Math(**_eval_script_returning_object('Math', as_kwargs=True))
    @Math.setter
    def Math(self, Math):
        raise AttributeError("Attribute 'Math' is read-only")

    @property
    def premierepro13(self):
        return _format_object_to_py(_eval_script_returning_object('premierepro13'))
    @premierepro13.setter
    def premierepro13(self, premierepro13):
        _eval_on_global_object("premierepro13 = {}".format(_format_object_to_es(premierepro13)))

    @property
    def AEFTBridge(self):
        return _format_object_to_py(_eval_script_returning_object('AEFTBridge'))
    @AEFTBridge.setter
    def AEFTBridge(self, AEFTBridge):
        _eval_on_global_object("AEFTBridge = {}".format(_format_object_to_es(AEFTBridge)))

    @property
    def PHXSBridge(self):
        return _format_object_to_py(_eval_script_returning_object('PHXSBridge'))
    @PHXSBridge.setter
    def PHXSBridge(self, PHXSBridge):
        _eval_on_global_object("PHXSBridge = {}".format(_format_object_to_es(PHXSBridge)))

    @property
    def CCXHostBridge(self):
        return _format_object_to_py(_eval_script_returning_object('CCXHostBridge'))
    @CCXHostBridge.setter
    def CCXHostBridge(self, CCXHostBridge):
        _eval_on_global_object("CCXHostBridge = {}".format(_format_object_to_es(CCXHostBridge)))

    @property
    def f(self):
        return File(**_eval_script_returning_object('f', as_kwargs=True))
    @f.setter
    def f(self, f):
        PymiereBaseObject._check_type(f, File, 'StartVars.f')
        _eval_on_global_object("f = {}".format(_format_object_to_es(f)))

    @property
    def JSON(self):
        return _format_object_to_py(_eval_script_returning_object('JSON'))
    @JSON.setter
    def JSON(self, JSON):
        _eval_on_global_object("JSON = {}".format(_format_object_to_es(JSON)))


    # ----- FUNCTIONS -----
    def encodeURI(self, text):
        """
        Encodes a string after RFC2396. Create an UTF-8 ASCII encoded version of this string. The string is converted into UTF-8. Every non-alphanumeric character is encoded as a percent escape character of the form %xx, where xx is the hex value of the character. After the conversion to UTF-8 encoding and escaping, it is guaranteed that the string does not contain characters codes greater than 127. The list of characters not to be encoded is -_.!~'();/?:@&=+Dollar,#. The method returns false on errors. 
        :param text: The text to encode.
        :type text: str
        """
        PymiereBaseObject._check_type(text, str, 'arg "text" of function "StartVars.encodeURI"')
        return _eval_on_global_object("encodeURI({})".format(_format_object_to_es(text)))

    def encodeURIComponent(self, text):
        """
        Encodes a string after RFC2396. Create an UTF-8 ASCII encoded version of this string. The string is converted into UTF-8. Every non-alphanumeric character is encoded as a percent escape character of the form %xx, where xx is the hex value of the character. After the conversion to UTF-8 encoding and escaping, it is guaranteed that the string does not contain characters codes greater than 127. The list of characters not to be encoded is -_.!~'(). The method returns false on errors. 
        :param text: The text to encode.
        :type text: str
        """
        PymiereBaseObject._check_type(text, str, 'arg "text" of function "StartVars.encodeURIComponent"')
        return _eval_on_global_object("encodeURIComponent({})".format(_format_object_to_es(text)))

    def decodeURI(self, uri):
        """
        Decodes a string created with encodeURI(). 
        :param uri: The text to decode.
        :type uri: str
        """
        PymiereBaseObject._check_type(uri, str, 'arg "uri" of function "StartVars.decodeURI"')
        return _eval_on_global_object("decodeURI({})".format(_format_object_to_es(uri)))

    def decodeURIComponent(self, uri):
        """
        Decodes a string created with encodeURIComponent(). 
        :param uri: The text to decode.
        :type uri: str
        """
        PymiereBaseObject._check_type(uri, str, 'arg "uri" of function "StartVars.decodeURIComponent"')
        return _eval_on_global_object("decodeURIComponent({})".format(_format_object_to_es(uri)))

    def escape(self, text):
        """
        Creates a URL-encoded string from aString. In the new string, characters of aString that require URL encoding are replaced with the format %xx, where xx is the hexadecimal value of the character code in the Unicode character set.This format is used to transmit information appended to a URL during, for example, execution of the GET method.Use the unescape() global function to translate the string back into its original format. Returns a string which is aString URL-encoded. 
        :param aString: The string to be encoded.
        :type text: str
        """
        PymiereBaseObject._check_type(text, str, 'arg "text" of function "StartVars.escape"')
        return _eval_on_global_object("escape({})".format(_format_object_to_es(text)))

    def eval(self, source):
        """
        Evaluates its argument as a JavaScript script, and returns the result of evaluation. You can pass the result of an object's toSource() method to reconstruct that object. 
        :param stringExpression: The string to evaluate.
        :type source: str
        """
        PymiereBaseObject._check_type(source, str, 'arg "source" of function "StartVars.eval"')
        return _eval_on_global_object("eval({})".format(_format_object_to_es(source)))

    def uneval(self, what):
        """
        Creates a source code representation of the supplied argument, and returns it as a string. 
        :param what: The object to uneval.
        :type what: any
        """
        PymiereBaseObject._check_type(what, any, 'arg "what" of function "StartVars.uneval"')
        return _eval_on_global_object("uneval({})".format(_format_object_to_es(what)))

    def isFinite(self, what):
        """
        Evaluates an expression and reports whether the result is a finite number. Returns true if the expression is a finite number, false otherwise. False if the value is infinity or negative infinity. 
        :param expression: Any valid JavaScript expression.
        :type what: float
        """
        PymiereBaseObject._check_type(what, float, 'arg "what" of function "StartVars.isFinite"')
        return _eval_on_global_object("isFinite({})".format(_format_object_to_es(what)))

    def isNaN(self, what):
        """
        Evaluates an expression and reports whether the result is "Not-a-Number" (NaN). Returns true if the result of evaluation is not a number (NaN), false if the value is a number. 
        :param expression: Any valid JavaScript expression.
        :type what: float
        """
        PymiereBaseObject._check_type(what, float, 'arg "what" of function "StartVars.isNaN"')
        return _eval_on_global_object("isNaN({})".format(_format_object_to_es(what)))

    def parseInt(self, text, base):
        """
        Extracts an integer from a string. Parses a string to find the first set of characters, in a specified base, that can be converted to an integer, and returns that integer, or NaN if it does not encounter characters that it can convert to a number. 
        :param text: The string from which to extract an integer.
        :param base: The base of the string to parse (from base 2 to base 36). If not supplied, base is determined by the format of string.
        :type text: str
        :type base: float
        """
        PymiereBaseObject._check_type(text, str, 'arg "text" of function "StartVars.parseInt"')
        PymiereBaseObject._check_type(base, float, 'arg "base" of function "StartVars.parseInt"')
        return _eval_on_global_object("parseInt({}, {})".format(_format_object_to_es(text), _format_object_to_es(base)))

    def parseFloat(self, txt):
        """
        Extracts a floating-point number from a string. Parses a string to find the first set of characters that can be converted to a floating point number, and returns that number, or NaN if it does not encounter characters that it can converted to a number.The function supports exponential notation. 
        :param text: The string from which to extract a floating point number.
        :type txt: str
        """
        PymiereBaseObject._check_type(txt, str, 'arg "txt" of function "StartVars.parseFloat"')
        return _eval_on_global_object("parseFloat({})".format(_format_object_to_es(txt)))

    def unescape(self, uri):
        """
        Translates URL-encoded string into a regular string, and returns that string. Use the escape() global function to URL-encode strings. 
        :param stringExpression: The URL-encoded string to convert.
        :type uri: str
        """
        PymiereBaseObject._check_type(uri, str, 'arg "uri" of function "StartVars.unescape"')
        return _eval_on_global_object("unescape({})".format(_format_object_to_es(uri)))

    def localize(self, what):
        """
        Localizes a ZString-encoded string and merges additional arguments into the string. 
        :param what: The string to localize. A ZString-encoded string that can contain placeholder for additional arguments in the form %1 to %n.
        :param arguments: Optional argument(s) to be merged into the string. There may be more than one argument.
        :type what: any
        """
        PymiereBaseObject._check_type(what, any, 'arg "what" of function "StartVars.localize"')
        return _eval_on_global_object("localize({})".format(_format_object_to_es(what)))

    def isXMLName(self, name):
        """
        Returns true if the supplied string is a valid XML name. 
        :param name: The XML name to test.
        :type name: str
        """
        PymiereBaseObject._check_type(name, str, 'arg "name" of function "StartVars.isXMLName"')
        return _eval_on_global_object("isXMLName({})".format(_format_object_to_es(name)))

    def setDefaultXMLNamespace(self, ns):
        """
        Defines the default XML namespace. This is a replacement function for the standard JavaScript statement set default xml namespace. 
        :param namespace: The namespace to use. Omit this parameter to return to the empty namespace. This is either a Namespace object or a string.
        :type ns: Namespace
        """
        return _eval_on_global_object("setDefaultXMLNamespace({})".format(_format_object_to_es(ns)))

    def alert(self, prompt):
        """
        Displays an alert box 
        :param message: The text to display
        :param title: The title of the alert; ignored on the Macintosh
        :param errorIcon: Display an Error icon; ignored on the Macintosh
        :type prompt: str
        """
        PymiereBaseObject._check_type(prompt, str, 'arg "prompt" of function "StartVars.alert"')
        _eval_on_global_object("alert({})".format(_format_object_to_es(prompt)))

    def confirm(self, prompt):
        """
        Displays an alert box with Yes and No buttons; returns true for Yes 
        :param message: The text to display
        :param noAsDefault: Set to true to set the No button as the default button
        :param title: The title of the alert; ignored on the Macintosh
        :type prompt: str
        """
        PymiereBaseObject._check_type(prompt, str, 'arg "prompt" of function "StartVars.confirm"')
        return _eval_on_global_object("confirm({})".format(_format_object_to_es(prompt)))

    def prompt(self, prompt):
        """
        Displays a dialog allowing the user to enter text Returns null if the user cancelled the dialog, the text otherwise 
        :param prompt: The text to display
        :param default_: The default text to preset the edit field with
        :param title: The title of the dialog;
        :type prompt: str
        """
        PymiereBaseObject._check_type(prompt, str, 'arg "prompt" of function "StartVars.prompt"')
        return _eval_on_global_object("prompt({})".format(_format_object_to_es(prompt)))



"""
Collection of higher level functions using low level pymiere code
"""
import os
import platform
import pymiere
from pymiere.core import get_premiere_version

# premiere uses ticks as its base time unit, this is used to convert from ticks to seconds
TICKS_PER_SECONDS = 254016000000


def check_active_sequence(crash=True):
    """
    Check if a project is opened and if a sequence is active

    :param crash: (bool) crash or return status as bool
    :return: (bool), (bool) project is opened, sequence is active
    """
    # is there a project opened ?
    if not pymiere.objects.app.isDocumentOpen():
        msg = "No project is opened"
        if crash:
            raise IOError(msg)
        print(msg)
        return False, False
    # is there a sequence opened ?
    if not pymiere.objects.app.project.activeSequence:
        msg = "No sequence is currently opened in the UI"
        if crash:
            raise AttributeError(msg)
        print(msg)
        return True, False
    return True, True


def get_item_recursive(item, add_root=False, filter_function=lambda i: True):
    """
    Recursively browse the project items returning a list of item with associated filepaths

    :param item: (ProjectItem) start item to browse from
    :param add_root: (bool) add the root (first given item) to the list and to the path
    :param filter_function: (None or func) function applied on all item, should return a bool adding or dismissing it
    :return: (list of ProjectItem)
    """
    all_children = [item] if filter_function(item) and add_root else list()
    # no children under item
    if item.children is None:
        return all_children
    # iter over children
    for c in item.children:
        all_children.extend(get_item_recursive(c, add_root=True, filter_function=filter_function))
    return all_children


def list_sequences():
    """
    List available sequences in the opened project, print infos and return objects

    :return: (list of Sequence) all sequences available in the project
    """
    sequences = pymiere.objects.app.project.sequences
    print("Found {} sequences in project '{}'\n".format(len(sequences), pymiere.objects.app.project.name))
    for sequence in sequences:
        print("Name : '{}'\nPath : '{}'".format(sequence.name, sequence.projectItem.treePath))
        print("Sequence setting :")
        settings = sequence.getSettings()
        for setting in ["videoFrameRate", "videoFrameWidth", "videoFrameHeight", "videoPixelAspectRatio", "audioSampleRate"]:
            value = settings.__getattribute__(setting)
            if isinstance(value, pymiere.Time):
                value = 1/value.seconds
            print("\t{} : {}".format(setting, value))
        print("\n")
    return sequences


def list_video(sequence):
    """
    List all video clips on all tracks on this sequence

    :param sequence: (Sequence)
    :return: (list of Clip)
    """
    video_clips = list()
    tracks = sequence.videoTracks
    for track in tracks:
        print("Track :", track.name or track.id)  # Premiere 2017 doesn't have 'name' property on tracks
        clips = track.clips
        for clip in clips:
            print("\tName: {}".format(clip.name))  # Premiere 2017 doesn't have 'name' property on clips
            print("\t- {:.<12}{}".format("Path", clip.projectItem.getMediaPath() if not clip.isMGT() else "[No path, this is a Motion Graphics template]"))
            print("\t- {:.<12}{}".format("Start", clip.start.seconds))
            print("\t- {:.<12}{}".format("End", clip.end.seconds))
            print("\t- {:.<12}{}".format("In", clip.inPoint.seconds))
            print("\t- {:.<12}{}".format("Out", clip.outPoint.seconds))
            print("\t- {:.<12}{}".format("Duration", clip.duration.seconds))
            print("")
            video_clips.append(clip)
        print("")
    return video_clips


def edit_clip(clip, start_on_timeline, end_on_timeline, in_point_on_clip, out_point_on_clip, fps=None):
    """
    Trim or move clip. This in not available in ppro 2017, use insertClip or overwriteClip

    :param clip: (Clip)
    :param start_on_timeline: (int) Frame at which the clip should start in the sequence timeline
    :param end_on_timeline: (int) Frame at which the clip should end in the sequence timeline
    :param in_point_on_clip: (int) First frame we will see of the clip (in the clip own timeline)
    :param out_point_on_clip: (int) Last frame we will see of the clip (in the clip own timeline)
    """
    # get clip fps if no fps given
    if fps is None:
        fps = clip.projectItem.getFootageInterpretation().frameRate  # not available in ppro 2017, use sequence.timebase
    # check length match
    if end_on_timeline - start_on_timeline != out_point_on_clip - in_point_on_clip:
        raise ValueError("Duration in timeline ({} frames) doesn't match duration of clip ({} frames)".format(end_on_timeline - start_on_timeline, out_point_on_clip - in_point_on_clip))
    # set actual properties
    clip.end = time_from_seconds(end_on_timeline / fps)
    clip.start = time_from_seconds(start_on_timeline / fps)
    clip.inPoint = time_from_seconds(in_point_on_clip / fps)
    clip.outPoint = time_from_seconds(out_point_on_clip / fps)


def move_clip(clip, seconds):
    """
    Move a clip by the number of seconds. Negative values will move the clip left (before the original position),
    positive values move the clip right (after the original position)
    :param clip: (Clip) clip object we want to move
    :param seconds: (float) how many seconds we want to move the clip by, negative value will move left, positive right
    """
    params = ["end", "start"] if seconds > 0 else ["start", "end"]
    for param in params:
        time_object = getattr(clip, param)
        time_object.seconds = time_object.seconds + seconds
        setattr(clip, param, time_object)


def animate_effect_using_function(clip, effect_name, property_name, anim_func, overwrite=True, keyframe_per_seconds=1/25):
    """
    Reproduce the animation of the given function on an effect property
    :param clip: (Clip) clip object on which we want to change animate the effect
    :param effect_name: (str) name of the effect (ex: 'twirl')
    :param property_name: (str) name of the property on the effect (ex: 'angle')
    :param anim_func: (function) function that take a float number as input (seconds) and return the value for the effect at this time as float
    (ex: lambda s: math.cos(s*2)*10 for a sine wave, lambda s: int(s % 2)*10 for a blinking effect, ...)
    :param overwrite: (bool) if True, clear existing keys else only add keys
    :param keyframe_per_seconds: (float) number of keyframes to place each seconds
    """
    # filter effect by name
    effects = clip.components
    for effect in effects:
        if effect.displayName.lower() != effect_name.lower():
            continue
        # filter effect property by name
        for property in effect.properties:
            if property.displayName.lower() != property_name.lower():
                continue
            # check keyframes are supported
            if not property.areKeyframesSupported():
                raise ValueError("Keyframes are not supported on property '{}' of effect '{}' for clip '{}'".format(property.displayName, effect.displayName, clip.name))
            # activate keyframes on proerty (= check the stop watch icon in the ui)
            if not property.isTimeVarying():
                property.setTimeVarying(True)

            # get clip in/out
            start = clip.inPoint.seconds
            end = clip.outPoint.seconds

            # clear keys
            if overwrite:
                property.removeKeyRange(start, end, True)

            # place keys
            while start < end:
                property.addKey(start)
                property.setValueAtKey(start, anim_func(start), True)
                start += keyframe_per_seconds


def time_from_seconds(seconds):
    """
    Many properties need a Time object to represent time, this is an helper to create one
    :param seconds: (float) time in seconds
    :return: (pymiere.Time) preseted time object
    """
    t = pymiere.Time()
    t.seconds = seconds
    return t


def get_system_sequence_presets(category="Digital SLR", resolution="1080p", preset_name="DSLR 1080p25"):
    """
    To create a new sequence via qe.project.newSequence we need to give a sequence preset file (.sqpreset)
    Base presets come installed with premiere. Select one according to your footage.
    Paths examples : (versions may vary)
        on win: C:\Program Files\Adobe\Adobe Premiere Pro 2020\Settings\SequencePresets
        on mac: /Applications/Adobe Premiere Pro CC 2019/Adobe Premiere Pro CC 2019.app/Contents/Settings/SequencePresets
    :param category: (str) category of the sequence preset in SequencePresets folder (AVCHD, RED R3D, ProRes RAW...)
    :param resolution: (str) category subfolder for sequence resolution
    :param preset_name: (str) actual filename of the sqpreset file (with or without extension)
    :return: (str) path of a sqpreset file
    """
    # add ext if needed
    if not preset_name.endswith(".sqpreset"):
        preset_name += ".sqpreset"
    # relative path
    if resolution is None:
        sequence_preset_root = os.path.normpath("Settings/SequencePresets/{}/{}".format(category, preset_name))
    else:
        sequence_preset_root = os.path.normpath("Settings/SequencePresets/{}/{}/{}".format(category, resolution, preset_name))
    # on mac add a Contents folder
    if platform.system().lower() != "windows":
        sequence_preset_root = os.path.join("Contents", sequence_preset_root)
    # absolute path in running Premiere pro directory
    sequence_preset_path = os.path.join(pymiere.objects.app.path, sequence_preset_root)
    # check exists
    if not os.path.isfile(sequence_preset_path):
        raise IOError("Sequence preset '{}' not found on disk".format(sequence_preset_path))
    return sequence_preset_path


def has_media_encoder():
    """
    Check if there is a Media Encoder (AME) version installed that work with the currently running Premiere Pro

    :return: (bool) True if suitable AME version found
    """
    premiere_version = get_premiere_version().version[0]
    encoder_versions = [int(app_name.split("-")[1].split(".")[0]) for app_name in pymiere.objects.apps if app_name.startswith("ame-")]
    return premiere_version in encoder_versions


def clone_sequence(sequence, new_sequence_name=None):
    """
    Equivalent of Sequence.clone() with return value and optional rename
    :param sequence: (Sequence) sequence object to duplicate
    :param new_sequence_name: (str or None) optional new name for the duplicated sequence
    :return: (Sequence) duplicated sequence object
    """
    project = pymiere.objects.app.project
    # store sequences before clone
    before_sequence_names = [s.sequenceID for s in project.sequences]
    # actual clone
    sequence.clone()
    # search clone sequence
    for new_seq in project.sequences:
        if new_seq.sequenceID not in before_sequence_names:
            break
    else:
        raise ValueError("New sequence not found")
    # rename and return
    if new_sequence_name is not None:
        new_seq.name = new_sequence_name
    return new_seq


if __name__ == "__main__":
    pass



import os
from setuptools import setup, find_packages

NAME = "pymiere"
URL = 'https://github.com/qmasingarbe/{}'.format(NAME)
VERSION = "1.2.6"

with open(os.path.join(os.path.dirname(__file__), "README.md"), "r") as f:
    long_description = f.read()

setup(name=NAME,
      packages=find_packages(),
      version=VERSION,
      license='GNU',
      description="Pythonic automations for Adobe Premiere Pro. Require the `Pymiere Link` extension for Premiere (installed separately).",
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Quentin Masingarbe',
      author_email='q.masingarbe@gmail.com',
      url=URL,
      keywords=[
          NAME, "video", "Premiere", "Adobe", "Pymiere", "workflow", "automation", "Creative Cloud", "edit",
          "editor", "editing", "prproj", "premiere pro"
      ],
      install_requires=['requests>=2.25'],
      # https://pypi.org/classifiers/
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                   'Natural Language :: English',
                   'Operating System :: Microsoft :: Windows',
                   'Operating System :: MacOS',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7',
                   'Programming Language :: Python :: 3.8',
                   'Programming Language :: Python :: 3.9',
                   'Topic :: Multimedia :: Video',
                   'Topic :: Multimedia :: Video :: Non-Linear Editor']
)

# Update VERSION (above) then run the following from the command prompt:

# python -m pip install setuptools wheel twine
# python setup.py sdist
# python -m twine upload --repository pypi dist/*

# To test:
# pip install pymiere



"""
Suite of function used to manipulate Premiere Pro app (is it running ? start it...)
Tested on Windows 10 and macOS Catalinq
Should work fine but no guaranty is provided...
"""
import os
import sys
import time
import re
import json
import subprocess
from distutils.version import StrictVersion
import platform
if platform.system().lower() == "windows":
    WINDOWS_SYSTEM = True
    try:
        import _winreg as wr  # python 2
    except:
        import winreg as wr  # python 3
else:
    # if not windows, assume it is a macOS
    WINDOWS_SYSTEM = False


CREATE_NO_WINDOW = 0x08000000
PREMIERE_PROCESS_NAME = "adobe premiere pro.exe" if WINDOWS_SYSTEM else "Adobe Premiere Pro"
CEPPANEL_PROCESS_NAME = "CEPHtmlEngine.exe" if WINDOWS_SYSTEM else "CEPHtmlEngine"


def is_premiere_running():
    """
    Is there a running instance of the Premiere Pro app on this machine ?

    :return: (bool) process is running, (int) pid
    """
    return exe_is_running(PREMIERE_PROCESS_NAME)


def start_premiere(use_bat=False):
    """
    Start Premiere pro if not already started

    :param use_bat: (bool) start Premiere Pro using a bat file to keep it running after script exit (in specific cases, windows only)
    :return (int) pid of Premiere process
    """
    running, pid = is_premiere_running()
    if running:
        return pid

    exe_path = get_last_premiere_exe()
    # we count the CEP pannel process running before because Premiere pops new ones at the end of loading
    start_running_cep_pannels = count_running_exe(CEPPANEL_PROCESS_NAME)

    if use_bat and WINDOWS_SYSTEM:
        # we don't call directly premiere exec here so it's not a child of this script.
        # It will still run after this script is killed (seems to be always the case except in Pycharm...)
        subprocess.call([os.path.join(__file__, "..", "bin", "start_premiere.bat"), exe_path])
    elif WINDOWS_SYSTEM:
        subprocess.Popen(exe_path, creationflags=subprocess.CREATE_NEW_CONSOLE)
    else:
        subprocess.Popen(["open", exe_path])

    # check process is starting and when it is done
    for i in range(200):
        time.sleep(0.5)
        running, pid = exe_is_running(PREMIERE_PROCESS_NAME)
        if not running:
            raise ValueError("Could not start premiere")
        current_running_cep_pannels = count_running_exe(CEPPANEL_PROCESS_NAME)
        if current_running_cep_pannels > start_running_cep_pannels:
            time.sleep(1)
            return pid
    raise SystemError("Could not guaranty premiere started")


def exe_is_running(exe_name):
    """
    List processes by name to know if one is running

    :param exe_name: (str) exact name of the process (ex : 'pycharm64.exe' for windows or 'Safari' for mac)
    :return: (bool) process is running, (int) pid
    """
    pids = _get_pids_from_name(exe_name)
    if len(pids) == 0:
        return False, None
    if len(pids) > 1:
        raise OSError("More than one process matching name '{}' were found running (pid: {})".format(exe_name, pids))
    return True, pids[0]


def count_running_exe(exe_name):
    """
    List processes by name to know how many are running

    :param exe_name: (str) exact name of the process (ex : 'pycharm64.exe' for windows or 'Safari' for mac)
    :return: (int) Number of process with given name running
    """
    return len(_get_pids_from_name(exe_name))


def get_last_premiere_exe():
    """
    Get the executable path on disk of the last installed Premiere Pro version

    :return: (str) path to executable
    """
    get_last_premiere_exe_func = _get_last_premiere_exe_windows if WINDOWS_SYSTEM else _get_last_premiere_exe_mac
    return get_last_premiere_exe_func()


def _get_pids_from_name(process_name):
    """
    Given a process name get ids of running process matching this name

    :param process_name: (str) process name (ex : 'pycharm64.exe' for windows or 'Safari' for mac)
    :return: (list of int) pids
    """
    if WINDOWS_SYSTEM:
        # use tasklist windows command with filter by name
        call = 'TASKLIST', '/FI', 'imagename eq {}'.format(process_name)
        output = subprocess.check_output(call, creationflags=CREATE_NO_WINDOW)
        if sys.version_info >= (3, 0):
            output = output.decode(encoding="437")  # encoding for windows console
        # parse output lines
        lines = output.strip().splitlines()
        matching_lines = [l for l in lines if l.lower().startswith(process_name.lower())]
        return [int(re.findall("   ([0-9]{1,6}) [a-zA-Z]", l)[0]) for l in matching_lines]
    else:
        # use pgrep UNIX command to filter processes by name
        try:
            output = subprocess.check_output(["pgrep", process_name])
        except subprocess.CalledProcessError:  # pgrep seems to crash if the given name is not a running process...
            return list()
        # parse output lines
        lines = output.strip().splitlines()
        return list(map(int, lines))


# ----- platform specific functions -----
def _get_last_premiere_exe_windows():
    """
    WINDOWS ONLY
    Get the executable path on disk of the last installed Premiere Pro version using windows registry

    :return: (str) path to executable
    """
    premiere_versions = _get_installed_softwares_info("adobe premiere pro")
    if not premiere_versions:
        raise OSError("Could not find an Adobe Premiere Pro version installed on this computer")
    # find last installed version
    last_version_num = sorted([StrictVersion(v["DisplayVersion"]) for v in premiere_versions])[-1]
    last_version_info = [v for v in premiere_versions if v["DisplayVersion"] == str(last_version_num)][0]
    # search actual exe path
    base_path = last_version_info["InstallLocation"]
    build_year = last_version_info["DisplayName"].split(" ")[-1]
    wrong_paths = list()
    for folder_name in ["Adobe Premiere Pro CC {}", "Adobe Premiere Pro {}", ""]:  # different versions formatting
        exe_path = os.path.join(base_path, folder_name.format(build_year), "Adobe Premiere Pro.exe")
        if not os.path.isfile(exe_path):
            wrong_paths.append(exe_path)
            continue
        wrong_paths = list()
        break
    if len(wrong_paths) != 0:
        raise IOError("Could not find Premiere executable in '{}'".format(wrong_paths))
    return exe_path


def _get_last_premiere_exe_mac():
    """
    MACOS ONLY
    Get the executable path on disk of the last installed Premiere Pro version using macOS System Profiler

    :return: (str) path to executable
    """
    # list all installed app to a json datastructure
    output = subprocess.check_output(["system_profiler", "-json", "SPApplicationsDataType"])
    apps_data = json.loads(output)["SPApplicationsDataType"]
    # filter Premiere pro installed versions
    premiere_apps = [data for data in apps_data if "adobe premiere pro" in data["_name"].lower()]
    if not premiere_apps:
        raise OSError("Could not find an Adobe Premiere Pro version installed on this computer")
    # get last app version path
    premiere_apps.sort(key=lambda d: d["version"], reverse=True)
    return premiere_apps[0]["path"]


def _get_installed_softwares_info(name_filter, names=["DisplayVersion", "InstallLocation"]):
    """
    WINDOWS ONLY
    Looking into Uninstall key in Windows registry, we can get some infos about installed software

    :param name_filter: (str) filter software containing this name
    :return: (list of dict) info of software found
    """
    reg = wr.ConnectRegistry(None, wr.HKEY_LOCAL_MACHINE)
    key = wr.OpenKey(reg, r"SOFTWARE\Wow6432Node\Microsoft\Windows\CurrentVersion\Uninstall")
    apps_info = list()
    # list all installed apps
    for i in range(wr.QueryInfoKey(key)[0]):
        subkey_name = wr.EnumKey(key,i)
        subkey = wr.OpenKey(key, subkey_name)
        try:
            soft_name = wr.QueryValueEx(subkey, "DisplayName")[0]
        except EnvironmentError:
            continue
        if name_filter.lower() not in soft_name.lower():
            continue
        apps_info.append(dict({n: wr.QueryValueEx(subkey, n)[0] for n in names}, DisplayName=soft_name))
    return apps_info


if __name__ == "__main__":
    # print(get_installed_softwares_info("adobe premiere pro"))
    # print(exe_is_running("adobe premiere pro.exe"))
    print(start_premiere())
